"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Runner = void 0;
const firebase_admin_1 = __importDefault(require("firebase-admin"));
const chalk_1 = __importDefault(require("chalk"));
const types_1 = require("./types");
const utils_1 = require("./utils");
const firestore_1 = require("firebase-admin/lib/firestore");
const assert = __importStar(require("assert"));
var Timestamp = firestore_1.firestore.Timestamp;
const isObject = (value) => 'object' === typeof value && value !== null;
const isString = (value) => 'string' === typeof value;
const isServiceAccountCredential = (credential) => isObject(credential) && isString(credential.projectId);
const puts = (value) => console.log(value);
const printProjectId = (app) => puts(`Using: ${chalk_1.default.bold(isServiceAccountCredential(app.options.credential) ? app.options.credential.projectId : 'Unknown')}\n`);
const teardownError = (error) => {
    console.error(error);
    process.exit(1);
};
const teardown = (fn) => async (...args) => Promise.resolve(fn(...args)).catch(teardownError);
class Runner {
    constructor(options) {
        this.app = firebase_admin_1.default.initializeApp();
        this.firestore = this.app.firestore();
        this.migrationCollection = this.firestore.collection(options.collectionName || 'migrations');
        this.migrationsDir = options.path;
    }
    static instance(options) {
        return new Runner(options);
    }
    async migrate(migration) {
        const input = this.createMigrationInput();
        const teardownFn = teardown(async () => {
            if (utils_1.isIrreversible(migration)) {
                await migration.execute(input);
            }
            else {
                await migration.up(input);
            }
        });
        await teardownFn();
    }
    async revert(dryRun, force, searchString) {
        printProjectId(this.app);
        const foundFiles = utils_1.getMigrationsFiles({ migrationsDir: this.migrationsDir, searchString });
        const executedMigrations = await this.getExecutedMigrationLogs();
        const revertibleMigrations = [
            ...(searchString && force
                ? foundFiles
                : foundFiles.filter((migrationFileName) => !!executedMigrations.filter(([id]) => !!~migrationFileName.indexOf(id)).length)),
        ];
        const teardownFn = teardown(async (id, migration) => {
            await migration.down(this.createMigrationInput());
            await this.removeMigrationLog(id);
        });
        console.info(`Found ${revertibleMigrations.length} migrations to revert.`);
        for (const migrationFile of revertibleMigrations) {
            const [id] = utils_1.getMigrationInfo(migrationFile);
            const migration = await this.createMigration(migrationFile);
            assert.ok(!utils_1.isIrreversible(migration), `Cannot revert ${migrationFile} because it is an irreversible migration.`);
            if (!dryRun) {
                await teardownFn(id, migration);
            }
            console.info(`Migration ${migrationFile} reverted.`);
        }
    }
    async run(dryRun, force, searchString) {
        printProjectId(this.app);
        const foundFiles = utils_1.getMigrationsFiles({ migrationsDir: this.migrationsDir, searchString });
        const executedMigrations = await this.getExecutedMigrationLogs();
        const runnableMigrations = [
            ...(searchString && force
                ? foundFiles
                : foundFiles.filter((migrationFileName) => !executedMigrations.filter(([id]) => !!~migrationFileName.indexOf(id)).length)),
        ];
        console.info(`Found ${runnableMigrations.length} migrations to run.`);
        for (const migrationFile of runnableMigrations) {
            const [id, timestamp, name] = utils_1.getMigrationInfo(migrationFile);
            const migration = await this.createMigration(migrationFile);
            if (!dryRun) {
                const type = utils_1.isIrreversible(migration) ? types_1.MigrationType.irreversible : types_1.MigrationType.reversible;
                await this.migrate(migration);
                await this.saveMigrationLog(id, {
                    type,
                    timestamp,
                    name,
                    executedAt: new Date(),
                });
            }
            console.info(`Migration ${migrationFile} executed.`);
        }
    }
    async list(options) {
        printProjectId(this.app);
        const logMigrations = (list) => void console.log(list.join('\n'));
        if (options.all) {
            const list = utils_1.getMigrationsFiles({ migrationsDir: options.path });
            logMigrations(list);
            return;
        }
        if (options.executed) {
            const executed = await this.getExecutedMigrationLogs();
            const foundFiles = utils_1.getMigrationsFiles({ migrationsDir: options.path });
            const list = foundFiles.filter((migrationFileName) => !!executed.filter(([id]) => !!~migrationFileName.indexOf(id)).length);
            logMigrations(list);
            return;
        }
        const executed = await this.getExecutedMigrationLogs();
        const foundFiles = utils_1.getMigrationsFiles({ migrationsDir: options.path });
        const list = foundFiles.filter((migrationFileName) => !executed.filter(([id]) => !!~migrationFileName.indexOf(id)).length);
        logMigrations(list);
    }
    async createMigration(migrationFileName) {
        const migrationPath = utils_1.getAbsolutePath({ migrationsDir: this.migrationsDir, migrationFileName });
        const { default: MigrationClass } = await Promise.resolve().then(() => __importStar(require(migrationPath)));
        return new MigrationClass();
    }
    createMigrationInput() {
        return { firestore: this.firestore, app: this.app };
    }
    async saveMigrationLog(id, fields) {
        await this.migrationCollection.doc(id).set({
            ...fields,
            ...((fields === null || fields === void 0 ? void 0 : fields.executedAt) && { executedAt: Timestamp.fromDate(fields.executedAt) }),
        }, { merge: true });
    }
    async removeMigrationLog(id) {
        await this.migrationCollection.doc(id).delete();
    }
    async getExecutedMigrationLogs() {
        return this.migrationCollection
            .orderBy('timestamp', 'asc')
            .get()
            .then((snapshot) => snapshot.docs.map((doc) => [doc.id, doc.data()]));
    }
}
exports.Runner = Runner;
